@@ -65,6 +65,28 @@
  * need to be implemented - this type implements these methods..
  */
 
+struct _GoaOAuth2ProviderPrivate
+{
+  GtkDialog *dialog;
+  GError *error;
+  GMainLoop *loop;
+
+  WebKitDOMHTMLInputElement *password_node;
+  const gchar *existing_identity;
+
+  gchar *account_object_path;
+
+  gchar *authorization_code;
+  gchar *access_token;
+  gint   access_token_expires_in;
+  gchar *refresh_token;
+  gchar *identity;
+  gchar *presentation_identity;
+  gchar *password;
+};
+
+#define GOA_OAUTH2_PROVIDER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GOA_TYPE_OAUTH2_PROVIDER, GoaOAuth2ProviderPrivate))
+
 G_LOCK_DEFINE_STATIC (provider_lock);
 
 G_DEFINE_ABSTRACT_TYPE (GoaOAuth2Provider, goa_oauth2_provider, GOA_TYPE_PROVIDER);
@@ -508,6 +530,66 @@ goa_oauth2_provider_is_identity_node (GoaOAuth2Provider *provider, WebKitDOMHTML
   return GOA_OAUTH2_PROVIDER_GET_CLASS (provider)->is_identity_node (provider, element);
 }
 
+/* ---------------------------------------------------------------------------------------------------- */
+
+static SoupCookie*
+create_cookie_from_variant (GVariant *variant)
+{
+  SoupCookie *cookie;
+  gchar *name, *value, *domain, *path;
+  gboolean secure, http_only;
+
+  g_return_val_if_fail (g_variant_is_of_type (variant, G_VARIANT_TYPE_VARDICT), NULL);
+
+  g_variant_lookup (variant, "name", "&s", &name);
+  g_variant_lookup (variant, "value", "&s", &value);
+  g_variant_lookup (variant, "domain", "&s", &domain);
+  g_variant_lookup (variant, "path", "&s", &path);
+  g_variant_lookup (variant, "secure", "b", &secure);
+  g_variant_lookup (variant, "http_only", "b", &http_only);
+
+  cookie = soup_cookie_new (name, value, domain, path, -1);
+  soup_cookie_set_secure (cookie, secure);
+  soup_cookie_set_http_only (cookie, http_only);
+
+  return cookie;
+}
+
+static GSList*
+extract_cookies_from_preseed_data (GVariant *preseed)
+{
+  GVariant *cookies_v, *cookie_v;
+  GSList *cookies = NULL;
+  GVariantIter iter;
+
+  if (preseed == NULL)
+    return NULL;
+
+  cookies_v = g_variant_lookup_value (preseed, "cookies", NULL);
+  if (cookies_v == NULL)
+    return NULL;
+
+  if (!g_variant_is_of_type (cookies_v, G_VARIANT_TYPE ("av")))
+    {
+      g_warning ("Wrong type for the 'cookies' GVariant, expected 'av' but got '%s'",
+                 (gchar *)g_variant_get_type (cookies_v));
+      goto out;
+    }
+
+  g_variant_iter_init (&iter, cookies_v);
+  while ((cookie_v = g_variant_iter_next_value (&iter)) != NULL)
+    {
+      SoupCookie *cookie = create_cookie_from_variant (g_variant_get_variant (cookie_v));
+      if (cookie != NULL)
+          cookies = g_slist_prepend (cookies, cookie);
+      g_variant_unref(cookie_v);
+    }
+out:
+  g_variant_unref(cookies_v);
+  return cookies;
+}
+
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 static gchar *
@@ -672,35 +754,15 @@ get_tokens_sync (GoaOAuth2Provider  *provider,
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
-typedef struct
-{
-  GoaOAuth2Provider *provider;
-  GtkDialog *dialog;
-  GError *error;
-  GMainLoop *loop;
-
-  WebKitDOMHTMLInputElement *password_node;
-  gchar *password;
-
-  gchar *authorization_code;
-  gchar *access_token;
-  gint access_token_expires_in;
-  gchar *refresh_token;
-
-  const gchar *existing_identity;
-
-  gchar *identity;
-  gchar *presentation_identity;
-} IdentifyData;
 
 static void
 check_cookie (SoupCookie *cookie, gpointer user_data)
 {
-  IdentifyData *data = user_data;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   GList *children;
   GList *l;
-  GoaOAuth2Provider *provider = data->provider;
-  GtkDialog *dialog = data->dialog;
+  GtkDialog *dialog = priv->dialog;
   GtkWidget *action_area;
   const gchar *auth_cookie;
   const gchar *name;
@@ -723,28 +785,30 @@ check_cookie (SoupCookie *cookie, gpointer user_data)
 static void
 on_dom_node_click (WebKitDOMNode *element, WebKitDOMEvent *event, gpointer user_data)
 {
-  IdentifyData *data = user_data;
-  gtk_dialog_response (data->dialog, GTK_RESPONSE_CANCEL);
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
+  gtk_dialog_response (priv->dialog, GTK_RESPONSE_CANCEL);
 }
 
 static void
 on_form_submit (WebKitDOMNode *element, WebKitDOMEvent *event, gpointer user_data)
 {
-  IdentifyData *data = user_data;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
 
-  if (data->password_node == NULL)
+  if (priv->password_node == NULL)
     return;
 
-  data->password = webkit_dom_html_input_element_get_value (data->password_node);
-  data->password_node = NULL;
+  priv->password = webkit_dom_html_input_element_get_value (priv->password_node);
+  priv->password_node = NULL;
 }
 
 static void
 on_web_view_document_load_finished (WebKitWebView *web_view, WebKitWebFrame *frame, gpointer user_data)
 {
-  IdentifyData *data = user_data;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   GSList *slist;
-  GoaOAuth2Provider *provider = data->provider;
   SoupCookieJar *cookie_jar;
   SoupSession *session;
   WebKitDOMDocument *document;
@@ -755,7 +819,7 @@ on_web_view_document_load_finished (WebKitWebView *web_view, WebKitWebFrame *fra
   session = webkit_get_default_session ();
   cookie_jar = SOUP_COOKIE_JAR (soup_session_get_feature (session, SOUP_TYPE_COOKIE_JAR));
   slist = soup_cookie_jar_all_cookies (cookie_jar);
-  g_slist_foreach (slist, (GFunc) check_cookie, data);
+  g_slist_foreach (slist, (GFunc) check_cookie, provider);
   g_slist_free_full (slist, (GDestroyNotify) soup_cookie_free);
 
   document = webkit_web_view_get_dom_document (WEBKIT_WEB_VIEW (web_view));
@@ -772,14 +836,14 @@ on_web_view_document_load_finished (WebKitWebView *web_view, WebKitWebFrame *fra
                                                       "click",
                                                       G_CALLBACK (on_dom_node_click),
                                                       FALSE,
-                                                      data);
+                                                      provider);
         }
-      else if (data->existing_identity != NULL
+      else if (priv->existing_identity != NULL
                && WEBKIT_DOM_IS_HTML_INPUT_ELEMENT (element)
                && goa_oauth2_provider_is_identity_node (provider, WEBKIT_DOM_HTML_INPUT_ELEMENT (element)))
         {
           webkit_dom_html_input_element_set_value (WEBKIT_DOM_HTML_INPUT_ELEMENT (element),
-                                                   data->existing_identity);
+                                                   priv->existing_identity);
           webkit_dom_html_input_element_set_read_only (WEBKIT_DOM_HTML_INPUT_ELEMENT (element), TRUE);
         }
       else if (WEBKIT_DOM_IS_HTML_INPUT_ELEMENT (element)
@@ -790,13 +854,13 @@ on_web_view_document_load_finished (WebKitWebView *web_view, WebKitWebFrame *fra
           form = webkit_dom_html_input_element_get_form (WEBKIT_DOM_HTML_INPUT_ELEMENT (element));
           if (form != NULL)
             {
-              data->password_node = WEBKIT_DOM_HTML_INPUT_ELEMENT (element);
-              g_clear_pointer (&data->password, g_free);
+              priv->password_node = WEBKIT_DOM_HTML_INPUT_ELEMENT (element);
+              g_clear_pointer (&priv->password, g_free);
               webkit_dom_event_target_add_event_listener (WEBKIT_DOM_EVENT_TARGET (form),
                                                           "submit",
                                                           G_CALLBACK (on_form_submit),
                                                           FALSE,
-                                                          data);
+                                                          provider);
             }
         }
     }
@@ -810,120 +874,119 @@ on_web_view_navigation_policy_decision_requested (WebKitWebView             *web
                                                   WebKitWebPolicyDecision   *policy_decision,
                                                   gpointer                   user_data)
 {
-  IdentifyData *data = user_data;
-  GHashTable *key_value_pairs;
-  SoupMessage *message;
-  SoupURI *uri;
-  const gchar *fragment;
-  const gchar *oauth2_error;
-  const gchar *query;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   const gchar *redirect_uri;
   const gchar *requested_uri;
-  gint response_id;
 
   /* TODO: use oauth2_proxy_extract_access_token() */
 
   requested_uri = webkit_network_request_get_uri (request);
-  redirect_uri = goa_oauth2_provider_get_redirect_uri (data->provider);
-  if (!g_str_has_prefix (requested_uri, redirect_uri))
-    goto default_behaviour;
-
-  message = webkit_network_request_get_message (request);
-  uri = soup_message_get_uri (message);
-  fragment = soup_uri_get_fragment (uri);
-  query = soup_uri_get_query (uri);
-
-  /* Two cases:
-   * 1) we can either have the access_token and other information
-   *    directly in the fragment part of the URI, or
-   * 2) the auth code can be in the query part of the URI, with which
-   *    we'll obtain the token later.
-   */
-  if (fragment != NULL)
+  redirect_uri = goa_oauth2_provider_get_redirect_uri (provider);
+  if (g_str_has_prefix (requested_uri, redirect_uri))
     {
-      /* fragment is encoded into a key/value pairs for the token and
-       * expiration values, using the same syntax as a URL query */
-      key_value_pairs = soup_form_decode (fragment);
+      SoupMessage *message;
+      SoupURI *uri;
 
-      /* We might use oauth2_proxy_extract_access_token() here but
-       * we can also extract other information.
+      message = webkit_network_request_get_message (request);
+      uri = soup_message_get_uri (message);
+
+      /* Two cases:
+       * 1) we can have either the auth code in the query part of the
+       *    URI, with which we'll obtain the token, or
+       * 2) the access_token and other information directly in the
+       *    fragment part of the URI.
        */
-      data->access_token = g_strdup (g_hash_table_lookup (key_value_pairs, "access_token"));
-      if (data->access_token != NULL)
+      if (soup_uri_get_query (uri) != NULL)
         {
-          gchar *expires_in_str = NULL;
+          GHashTable *key_value_pairs;
 
-          expires_in_str = g_hash_table_lookup (key_value_pairs, "expires_in");
-          /* sometimes "expires_in" appears as "expires" */
-          if (expires_in_str == NULL)
-            expires_in_str = g_hash_table_lookup (key_value_pairs, "expires");
+          key_value_pairs = soup_form_decode (uri->query);
 
-          if (expires_in_str != NULL)
-            data->access_token_expires_in = atoi (expires_in_str);
-
-          data->refresh_token = g_strdup (g_hash_table_lookup (key_value_pairs, "refresh_token"));
-
-          response_id = GTK_RESPONSE_OK;
+          priv->authorization_code = g_strdup (g_hash_table_lookup (key_value_pairs, "code"));
+          if (priv->authorization_code != NULL)
+            {
+              gtk_dialog_response (priv->dialog, GTK_RESPONSE_OK);
+            }
+          else
+            {
+              g_set_error (&priv->error,
+                           GOA_ERROR,
+                           GOA_ERROR_NOT_AUTHORIZED,
+                           _("Authorization response was \"%s\""),
+                           (const gchar *) g_hash_table_lookup (key_value_pairs, "error"));
+              gtk_dialog_response (priv->dialog, GTK_RESPONSE_CLOSE);
+            }
+          g_hash_table_unref (key_value_pairs);
+          webkit_web_policy_decision_ignore (policy_decision);
         }
-      g_hash_table_unref (key_value_pairs);
+      else if (soup_uri_get_fragment (uri) != NULL)
+        {
+          GHashTable *key_value_pairs;
+
+          /* fragment is encoded into a key/value pairs for the token and
+           * expiration values, using the same syntax as a URL query */
+          key_value_pairs = soup_form_decode (soup_uri_get_fragment (uri));
+
+          /* We might use oauth2_proxy_extract_access_token() here but
+           * we can also extract other information.
+           */
+          priv->access_token = g_strdup (g_hash_table_lookup (key_value_pairs, "access_token"));
+          if (priv->access_token != NULL)
+            {
+              gchar *expires_in_str = NULL;
+
+              expires_in_str = g_hash_table_lookup (key_value_pairs, "expires_in");
+              /* sometimes "expires_in" appears as "expires" */
+              if (expires_in_str == NULL)
+                expires_in_str = g_hash_table_lookup (key_value_pairs, "expires");
+
+              if (expires_in_str != NULL)
+                 priv->access_token_expires_in = atoi (expires_in_str);
+
+              priv->refresh_token = g_strdup (g_hash_table_lookup (key_value_pairs, "refresh_token"));
+
+              gtk_dialog_response (priv->dialog, GTK_RESPONSE_OK);
+            }
+          else
+            {
+              g_set_error (&priv->error,
+                           GOA_ERROR,
+                           GOA_ERROR_NOT_AUTHORIZED,
+                           _("Authorization response was \"%s\""),
+                           (const gchar *) g_hash_table_lookup (key_value_pairs, "error"));
+              gtk_dialog_response (priv->dialog, GTK_RESPONSE_CLOSE);
+            }
+          g_hash_table_unref (key_value_pairs);
+          webkit_web_policy_decision_ignore (policy_decision);
+        }
+      else
+        {
+          /* this actually means that something unexpected happened, either we
+           * did something wrong or the provider's flow changed */
+          goa_debug ("URI format not recognized, DEFAULT BEHAVIOUR");
+          return FALSE;
+        }
+      return TRUE; /* ignore the request */
     }
-
-  if (data->access_token != NULL)
-    goto ignore_request;
-
-  if (query != NULL)
-    {
-      key_value_pairs = soup_form_decode (query);
-
-      data->authorization_code = g_strdup (g_hash_table_lookup (key_value_pairs, "code"));
-      if (data->authorization_code != NULL)
-        response_id = GTK_RESPONSE_OK;
-
-      g_hash_table_unref (key_value_pairs);
-    }
-
-  if (data->authorization_code != NULL)
-    goto ignore_request;
-
-  /* In case we don't find the access_token or auth code, then look
-   * for the error in the query part of the URI.
-   */
-  key_value_pairs = soup_form_decode (query);
-  oauth2_error = (const gchar *) g_hash_table_lookup (key_value_pairs, "error");
-  if (g_strcmp0 (oauth2_error, GOA_OAUTH2_ACCESS_DENIED) == 0)
-    response_id = GTK_RESPONSE_CANCEL;
   else
     {
-      g_set_error (&data->error,
-                   GOA_ERROR,
-                   GOA_ERROR_NOT_AUTHORIZED,
-                   _("Authorization response was \"%s\""),
-                   oauth2_error);
-      response_id = GTK_RESPONSE_CLOSE;
+      return FALSE; /* make default behavior apply */
     }
-  g_hash_table_unref (key_value_pairs);
-  goto ignore_request;
-
- ignore_request:
-  gtk_dialog_response (data->dialog, response_id);
-  webkit_web_policy_decision_ignore (policy_decision);
-  return TRUE;
-
- default_behaviour:
-  return FALSE;
 }
 
 static void
 on_entry_changed (GtkEditable *editable,
                   gpointer     user_data)
 {
-  IdentifyData *data = user_data;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   gboolean sensitive;
 
-  g_free (data->authorization_code);
-  data->authorization_code = g_strdup (gtk_entry_get_text (GTK_ENTRY (editable)));
-  sensitive = data->authorization_code != NULL && (strlen (data->authorization_code) > 0);
-  gtk_dialog_set_response_sensitive (data->dialog, GTK_RESPONSE_OK, sensitive);
+  g_free (priv->authorization_code);
+  priv->authorization_code = g_strdup (gtk_entry_get_text (GTK_ENTRY (editable)));
+  sensitive = priv->authorization_code != NULL && (strlen (priv->authorization_code) > 0);
+  gtk_dialog_set_response_sensitive (priv->dialog, GTK_RESPONSE_OK, sensitive);
 }
 
 static gboolean
@@ -931,20 +994,12 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
                          gboolean            add_account,
                          const gchar        *existing_identity,
                          GtkDialog          *dialog,
-                         GtkBox             *vbox,
-                         gchar             **out_authorization_code,
-                         gchar             **out_access_token,
-                         gint               *out_access_token_expires_in,
-                         gchar             **out_refresh_token,
-                         gchar             **out_identity,
-                         gchar             **out_presentation_identity,
-                         gchar             **out_password,
-                         GError            **error)
+                         GtkBox             *vbox)
 {
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   gboolean ret;
   gchar *url;
   GtkWidget *grid;
-  IdentifyData data;
   gchar *escaped_redirect_uri;
   gchar *escaped_client_id;
   gchar *escaped_scope;
@@ -954,7 +1009,7 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
                         || (add_account && existing_identity == NULL), FALSE);
   g_return_val_if_fail (GTK_IS_DIALOG (dialog), FALSE);
   g_return_val_if_fail (GTK_IS_BOX (vbox), FALSE);
-  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+  g_return_val_if_fail (priv->error == NULL, FALSE);
 
   ret = FALSE;
   escaped_redirect_uri = NULL;
@@ -963,11 +1018,15 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
 
   /* TODO: check with NM whether we're online, if not - return error */
 
-  memset (&data, '\0', sizeof (IdentifyData));
-  data.provider = provider;
-  data.dialog = dialog;
-  data.loop = g_main_loop_new (NULL, FALSE);
-  data.existing_identity = existing_identity;
+  priv->dialog = dialog;
+  priv->existing_identity = existing_identity;
+
+  g_clear_pointer (&priv->password, g_free);
+  g_clear_pointer (&priv->identity, g_free);
+  g_clear_pointer (&priv->presentation_identity, g_free);
+  g_clear_pointer (&priv->authorization_code, g_free);
+  g_clear_pointer (&priv->access_token, g_free);
+  g_clear_pointer (&priv->refresh_token, g_free);
 
   /* TODO: use oauth2_proxy_build_login_url_full() */
   escaped_redirect_uri = g_uri_escape_string (goa_oauth2_provider_get_redirect_uri (provider), NULL, TRUE);
@@ -1014,12 +1073,12 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
       gtk_dialog_add_button (dialog, GTK_STOCK_OK, GTK_RESPONSE_OK);
       gtk_dialog_set_default_response (dialog, GTK_RESPONSE_OK);
       gtk_dialog_set_response_sensitive (dialog, GTK_RESPONSE_OK, FALSE);
-      g_signal_connect (entry, "changed", G_CALLBACK (on_entry_changed), &data);
+      g_signal_connect (entry, "changed", G_CALLBACK (on_entry_changed), provider);
 
       if (!gtk_show_uri (NULL,
                          url,
                          GDK_CURRENT_TIME,
-                         &data.error))
+                         &priv->error))
         {
           goto out;
         }
@@ -1028,12 +1087,19 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
     {
       GtkWidget *web_view;
       GtkWidget *embed;
+      GVariant *preseed_data;
+      GSList *cookies;
 
       web_view = goa_web_view_new ();
       gtk_widget_set_hexpand (web_view, TRUE);
       gtk_widget_set_vexpand (web_view, TRUE);
       embed = goa_web_view_get_view (GOA_WEB_VIEW (web_view));
 
+      preseed_data = goa_provider_get_preseed_data (GOA_PROVIDER (provider));
+      cookies = extract_cookies_from_preseed_data (preseed_data);
+      goa_web_view_add_cookies (GOA_WEB_VIEW (web_view), cookies);
+      soup_cookies_free (cookies);
+
       if (goa_oauth2_provider_get_use_mobile_browser (provider))
         goa_web_view_fake_mobile (GOA_WEB_VIEW (web_view));
 
@@ -1041,11 +1107,11 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
       g_signal_connect (embed,
                         "document-load-finished",
                         G_CALLBACK (on_web_view_document_load_finished),
-                        &data);
+                        provider);
       g_signal_connect (embed,
                         "navigation-policy-decision-requested",
                         G_CALLBACK (on_web_view_navigation_policy_decision_requested),
-                        &data);
+                        provider);
 
       gtk_container_add (GTK_CONTAINER (grid), web_view);
     }
@@ -1057,22 +1123,22 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
    * the token directly if we are using a client side flow, since we don't
    * need to pass the code to the remote application.
    */
-  if (data.authorization_code == NULL && data.access_token == NULL)
+  if (priv->authorization_code == NULL && priv->access_token == NULL)
     {
-      if (data.error == NULL)
+      if (priv->error == NULL)
         {
-          g_set_error (&data.error,
+          g_set_error (&priv->error,
                        GOA_ERROR,
                        GOA_ERROR_DIALOG_DISMISSED,
                        _("Dialog was dismissed"));
         }
       goto out;
     }
-  g_assert (data.error == NULL);
+  g_assert (priv->error == NULL);
 
   gtk_widget_hide (GTK_WIDGET (dialog));
 
-  if (data.authorization_code != NULL)
+  if (priv->authorization_code != NULL)
     {
       /* OK, we now have the authorization code... now we need to get the
        * email address (to e.g. check if the account already exists on
@@ -1081,71 +1147,38 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
        */
 
       /* TODO: run in worker thread */
-      data.access_token = get_tokens_sync (provider,
-                                           data.authorization_code,
-                                           NULL, /* refresh_token */
-                                           &data.refresh_token,
-                                           &data.access_token_expires_in,
-                                           NULL, /* GCancellable */
-                                           &data.error);
-      if (data.access_token == NULL)
+      priv->access_token = get_tokens_sync (provider,
+                                            priv->authorization_code,
+                                            NULL, /* refresh_token */
+                                            &priv->refresh_token,
+                                            &priv->access_token_expires_in,
+                                            NULL, /* GCancellable */
+                                            &priv->error);
+      if (priv->access_token == NULL)
         {
-          g_prefix_error (&data.error, _("Error getting an Access Token: "));
+          g_prefix_error (&priv->error, _("Error getting an Access Token: "));
           goto out;
         }
     }
 
-  g_assert (data.access_token != NULL);
+  g_assert (priv->access_token != NULL);
 
   /* TODO: run in worker thread */
-  data.identity = goa_oauth2_provider_get_identity_sync (provider,
-                                                         data.access_token,
-                                                         &data.presentation_identity,
-                                                         NULL, /* TODO: GCancellable */
-                                                         &data.error);
-  if (data.identity == NULL)
+  priv->identity = goa_oauth2_provider_get_identity_sync (provider,
+                                                          priv->access_token,
+                                                          &priv->presentation_identity,
+                                                          NULL, /* TODO: GCancellable */
+                                                          &priv->error);
+  if (priv->identity == NULL)
     {
-      g_prefix_error (&data.error, _("Error getting identity: "));
+      g_prefix_error (&priv->error, _("Error getting identity: "));
       goto out;
     }
 
   ret = TRUE;
 
  out:
-  if (ret)
-    {
-      g_warn_if_fail (data.error == NULL);
-      if (out_authorization_code != NULL)
-        *out_authorization_code = g_strdup (data.authorization_code);
-      if (out_access_token != NULL)
-        *out_access_token = g_strdup (data.access_token);
-      if (out_access_token_expires_in != NULL)
-        *out_access_token_expires_in = data.access_token_expires_in;
-      if (out_refresh_token != NULL)
-        *out_refresh_token = g_strdup (data.refresh_token);
-      if (out_identity != NULL)
-        *out_identity = g_strdup (data.identity);
-      if (out_presentation_identity != NULL)
-        *out_presentation_identity = g_strdup (data.presentation_identity);
-      if (out_password != NULL)
-        *out_password = g_strdup (data.password);
-    }
-  else
-    {
-      g_warn_if_fail (data.error != NULL);
-      g_propagate_error (error, data.error);
-    }
-
-  g_free (data.password);
-  g_free (data.identity);
-  g_free (data.presentation_identity);
   g_free (url);
-
-  g_free (data.authorization_code);
-  if (data.loop != NULL)
-    g_main_loop_unref (data.loop);
-  g_free (data.access_token);
-  g_free (data.refresh_token);
   g_free (escaped_redirect_uri);
   g_free (escaped_client_id);
   g_free (escaped_scope);
@@ -1154,24 +1187,18 @@ get_tokens_and_identity (GoaOAuth2Provider  *provider,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-typedef struct
-{
-  GError *error;
-  GMainLoop *loop;
-  gchar *account_object_path;
-} AddData;
-
 static void
 add_account_cb (GoaManager   *manager,
                 GAsyncResult *res,
                 gpointer      user_data)
 {
-  AddData *data = user_data;
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (user_data);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   goa_manager_call_add_account_finish (manager,
-                                       &data->account_object_path,
+                                       &priv->account_object_path,
                                        res,
-                                       &data->error);
-  g_main_loop_quit (data->loop);
+                                       &priv->error);
+  g_main_loop_quit (priv->loop);
 }
 
 static gint64
@@ -1198,6 +1225,25 @@ abs_usec_to_duration (gint64 abs_usec)
   return ret;
 }
 
+static void
+add_credentials_key_values (GoaOAuth2Provider *provider,
+                            GVariantBuilder *credentials)
+{
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
+
+  if (priv->authorization_code != NULL)
+    g_variant_builder_add (credentials, "{sv}", "authorization_code",
+                           g_variant_new_string (priv->authorization_code));
+  g_variant_builder_add (credentials, "{sv}", "access_token", g_variant_new_string (priv->access_token));
+  if (priv->access_token_expires_in > 0)
+    g_variant_builder_add (credentials, "{sv}", "access_token_expires_at",
+                           g_variant_new_int64 (duration_to_abs_usec (priv->access_token_expires_in)));
+  if (priv->refresh_token != NULL)
+    g_variant_builder_add (credentials, "{sv}", "refresh_token", g_variant_new_string (priv->refresh_token));
+  if (priv->password != NULL)
+    g_variant_builder_add (credentials, "{sv}", "password", g_variant_new_string (priv->password));
+}
+
 static GoaObject *
 goa_oauth2_provider_add_account (GoaProvider *_provider,
                                          GoaClient          *client,
@@ -1206,15 +1252,8 @@ goa_oauth2_provider_add_account (GoaProvider *_provider,
                                          GError            **error)
 {
   GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (_provider);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   GoaObject *ret;
-  gchar *authorization_code;
-  gchar *access_token;
-  gint access_token_expires_in;
-  gchar *password;
-  gchar *refresh_token;
-  gchar *identity;
-  gchar *presentation_identity;
-  AddData data;
   GVariantBuilder credentials;
   GVariantBuilder details;
 
@@ -1225,53 +1264,27 @@ goa_oauth2_provider_add_account (GoaProvider *_provider,
   g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
   ret = NULL;
-  authorization_code = NULL;
-  access_token = NULL;
-  password = NULL;
-  refresh_token = NULL;
-  identity = NULL;
-  presentation_identity = NULL;
-
-  memset (&data, '\0', sizeof (AddData));
-  data.loop = g_main_loop_new (NULL, FALSE);
 
   if (!get_tokens_and_identity (provider,
                                 TRUE,
                                 NULL,
                                 dialog,
-                                vbox,
-                                &authorization_code,
-                                &access_token,
-                                &access_token_expires_in,
-                                &refresh_token,
-                                &identity,
-                                &presentation_identity,
-                                &password,
-                                &data.error))
+                                vbox))
     goto out;
 
   /* OK, got the identity... see if there's already an account
    * of this type with the given identity
    */
   if (!goa_utils_check_duplicate (client,
-                                  identity,
-                                  presentation_identity,
+                                  priv->identity,
+                                  priv->presentation_identity,
                                   goa_provider_get_provider_type (GOA_PROVIDER (provider)),
                                   (GoaPeekInterfaceFunc) goa_object_peek_oauth2_based,
-                                  &data.error))
+                                  &priv->error))
     goto out;
 
   g_variant_builder_init (&credentials, G_VARIANT_TYPE_VARDICT);
-  if (authorization_code != NULL)
-    g_variant_builder_add (&credentials, "{sv}", "authorization_code", g_variant_new_string (authorization_code));
-  g_variant_builder_add (&credentials, "{sv}", "access_token", g_variant_new_string (access_token));
-  if (access_token_expires_in > 0)
-    g_variant_builder_add (&credentials, "{sv}", "access_token_expires_at",
-                           g_variant_new_int64 (duration_to_abs_usec (access_token_expires_in)));
-  if (refresh_token != NULL)
-    g_variant_builder_add (&credentials, "{sv}", "refresh_token", g_variant_new_string (refresh_token));
-  if (password != NULL)
-    g_variant_builder_add (&credentials, "{sv}", "password", g_variant_new_string (password));
+  add_credentials_key_values (provider, &credentials);
 
   g_variant_builder_init (&details, G_VARIANT_TYPE ("a{ss}"));
   goa_oauth2_provider_add_account_key_values (provider, &details);
@@ -1282,38 +1295,32 @@ goa_oauth2_provider_add_account (GoaProvider *_provider,
    */
   goa_manager_call_add_account (goa_client_get_manager (client),
                                 goa_provider_get_provider_type (GOA_PROVIDER (provider)),
-                                identity,
-                                presentation_identity,
+                                priv->identity,
+                                priv->presentation_identity,
                                 g_variant_builder_end (&credentials),
                                 g_variant_builder_end (&details),
                                 NULL, /* GCancellable* */
                                 (GAsyncReadyCallback) add_account_cb,
-                                &data);
-  g_main_loop_run (data.loop);
-  if (data.error != NULL)
+                                provider);
+  priv->loop = g_main_loop_new (NULL, FALSE);
+  g_main_loop_run (priv->loop);
+  if (priv->error != NULL)
     goto out;
 
   ret = GOA_OBJECT (g_dbus_object_manager_get_object (goa_client_get_object_manager (client),
-                                                      data.account_object_path));
+                                                      priv->account_object_path));
 
  out:
-  /* We might have an object even when data.error is set.
+  /* We might have an object even when priv->error is set.
    * eg., if we failed to store the credentials in the keyring.
    */
-  if (data.error != NULL)
-    g_propagate_error (error, data.error);
+  if (priv->error != NULL)
+    g_propagate_error (error, priv->error);
   else
     g_assert (ret != NULL);
 
-  g_free (identity);
-  g_free (presentation_identity);
-  g_free (password);
-  g_free (refresh_token);
-  g_free (access_token);
-  g_free (authorization_code);
-  g_free (data.account_object_path);
-  if (data.loop != NULL)
-    g_main_loop_unref (data.loop);
+  g_clear_pointer (&priv->account_object_path, g_free);
+  g_clear_pointer (&priv->loop, g_main_loop_unref);
   return ret;
 }
 
@@ -1327,14 +1334,9 @@ goa_oauth2_provider_refresh_account (GoaProvider  *_provider,
                                      GError      **error)
 {
   GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (_provider);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
   GoaAccount *account;
   GtkWidget *dialog;
-  gchar *authorization_code;
-  gchar *access_token;
-  gint access_token_expires_in;
-  gchar *password;
-  gchar *refresh_token;
-  gchar *identity;
   const gchar *existing_identity;
   const gchar *existing_presentation_identity;
   GVariantBuilder builder;
@@ -1346,12 +1348,6 @@ goa_oauth2_provider_refresh_account (GoaProvider  *_provider,
   g_return_val_if_fail (parent == NULL || GTK_IS_WINDOW (parent), FALSE);
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-  authorization_code = NULL;
-  access_token = NULL;
-  password = NULL;
-  refresh_token = NULL;
-  identity = NULL;
-
   ret = FALSE;
 
   dialog = gtk_dialog_new_with_buttons (NULL,
@@ -1374,15 +1370,7 @@ goa_oauth2_provider_refresh_account (GoaProvider  *_provider,
                                 FALSE,
                                 existing_presentation_identity,
                                 GTK_DIALOG (dialog),
-                                GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
-                                &authorization_code,
-                                &access_token,
-                                &access_token_expires_in,
-                                &refresh_token,
-                                &identity,
-                                NULL, /* out_presentation_identity */
-                                &password,
-                                error))
+                                GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog)))))
     goto out;
 
   /* Changes made to the web interface by the providers can break our
@@ -1390,28 +1378,19 @@ goa_oauth2_provider_refresh_account (GoaProvider  *_provider,
    * afterwards.
    */
   existing_identity = goa_account_get_identity (account);
-  if (g_strcmp0 (identity, existing_identity) != 0)
+  if (g_strcmp0 (priv->identity, existing_identity) != 0)
     {
       g_set_error (error,
                    GOA_ERROR,
                    GOA_ERROR_FAILED,
                    _("Was asked to login as %s, but logged in as %s"),
                    existing_identity,
-                   identity);
+                   priv->identity);
       goto out;
     }
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE_VARDICT);
-  if (authorization_code != NULL)
-    g_variant_builder_add (&builder, "{sv}", "authorization_code", g_variant_new_string (authorization_code));
-  g_variant_builder_add (&builder, "{sv}", "access_token", g_variant_new_string (access_token));
-  if (access_token_expires_in > 0)
-    g_variant_builder_add (&builder, "{sv}", "access_token_expires_at",
-                           g_variant_new_int64 (duration_to_abs_usec (access_token_expires_in)));
-  if (refresh_token != NULL)
-    g_variant_builder_add (&builder, "{sv}", "refresh_token", g_variant_new_string (refresh_token));
-  if (password != NULL)
-    g_variant_builder_add (&builder, "{sv}", "password", g_variant_new_string (password));
+  add_credentials_key_values (provider, &builder);
   if (!goa_utils_store_credentials_for_object_sync (GOA_PROVIDER (provider),
                                                     object,
                                                     g_variant_builder_end (&builder),
@@ -1427,12 +1406,6 @@ goa_oauth2_provider_refresh_account (GoaProvider  *_provider,
 
  out:
   gtk_widget_destroy (dialog);
-
-  g_free (identity);
-  g_free (access_token);
-  g_free (authorization_code);
-  g_free (password);
-  g_free (refresh_token);
   return ret;
 }
 
@@ -1778,15 +1751,39 @@ goa_oauth2_provider_ensure_credentials_sync (GoaProvider   *_provider,
 /* ---------------------------------------------------------------------------------------------------- */
 
 static void
-goa_oauth2_provider_init (GoaOAuth2Provider *client)
+goa_oauth2_provider_finalize (GObject *object)
 {
+  GoaOAuth2Provider *provider = GOA_OAUTH2_PROVIDER (object);
+  GoaOAuth2ProviderPrivate *priv = provider->priv;
+
+  g_clear_pointer (&priv->loop, g_main_loop_unref);
+
+  g_free (priv->account_object_path);
+  g_free (priv->password);
+  g_free (priv->identity);
+  g_free (priv->presentation_identity);
+  g_free (priv->authorization_code);
+  g_free (priv->access_token);
+  g_free (priv->refresh_token);
+
+  G_OBJECT_CLASS (goa_oauth2_provider_parent_class)->finalize (object);
+}
+
+static void
+goa_oauth2_provider_init (GoaOAuth2Provider *provider)
+{
+  provider->priv = GOA_OAUTH2_PROVIDER_GET_PRIVATE (provider);
 }
 
 static void
 goa_oauth2_provider_class_init (GoaOAuth2ProviderClass *klass)
 {
+  GObjectClass *object_class;
   GoaProviderClass *provider_class;
 
+  object_class = G_OBJECT_CLASS (klass);
+  object_class->finalize = goa_oauth2_provider_finalize;
+
   provider_class = GOA_PROVIDER_CLASS (klass);
   provider_class->add_account                = goa_oauth2_provider_add_account;
   provider_class->refresh_account            = goa_oauth2_provider_refresh_account;
@@ -1799,6 +1796,8 @@ goa_oauth2_provider_class_init (GoaOAuth2ProviderClass *klass)
   klass->get_use_mobile_browser   = goa_oauth2_provider_get_use_mobile_browser_default;
   klass->is_password_node         = goa_oauth2_provider_is_password_node_default;
   klass->add_account_key_values   = goa_oauth2_provider_add_account_key_values_default;
+
+  g_type_class_add_private (object_class, sizeof (GoaOAuth2ProviderPrivate));
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
